-- =================================
--  Mahjong Solitair for Love2D
--  Programmed for Love2D by Toshi Nagata
--  Copyright 2019 (C) Toshi Nagata
--  Released under the MIT open source license.
--  tiles.png: Designed by Game Design, http://www.gamedesign.jp/
-- =================================

require "startup"

--  Layout of the tiles
--  (x1, y1) - (x2, y2), z
--  The positions should be in ascending order for z, y, x
layout = {
  { -6, -4, 2, -4, 0 },
  { -5, -3, 4, -3, 0 },
  { -4, -2, 5, -2, 0 },
  { -3, -1, 6, -1, 0 },
  { -2, 0, 7, 0, 0 },
  { -3, 1, 6, 1, 0 },
  { -4, 2, 5, 2, 0 },
  { -5, 3, 4, 3, 0 },
  { -2.5, -3, 2.5, -1, 1 },
  { -1.5, 0, 3.5, 2, 1 },
  { -1, -2, 2, 1, 2 },
  { 0, -1, 1, 0, 3 },
  { 0.5, -0.5, 0.5, -0.5, 4 }
}

--  Tiles
--  properties: x, y, z, id, flags
tiles = {}

--  Quads for displaying a single tile
quads = {}

--  取れる牌を種類ごとにカウント
count_movable = {}

--  取れる牌の総数
number_movable = nil

--  残っている牌の数
number_remain = nil

--  盤面の初期値
seed = 1

--  tiles のインデックスを y, x, z の昇順に並べ替えたもの
indices = {}

--  z 座標の最大値 (check_movable() でセット)
zmax = -10

--  カーソル位置 (indices の添字)
curs = 0

--  選択している牌 (indices の添字)
sel = 0

--  取れるかどうかを判定
function check_movable()
  local i, j, t1, t2, flag
  for i = 1, 34 do
    count_movable[i] = 0
  end
  zmax = -10
  for i = 1, 136 do
    t1 = tiles[i]
    if not t1.visible then goto cont end  -- すでに消えている
    flag = false
    for j = i + 1, 136 do
      t2 = tiles[j]
      if t2.visible and t2.z == t1.z + 1 and math.abs(t1.x - t2.x) < 1 and math.abs(t1.y - t2.y) < 1 then
        goto cont  -- 上に牌があるので動かせない
      end
    end
    if i == 1 then
      flag = true -- 動かせる
    else
      t2 = tiles[i - 1]
      if (not t2.visible) or t2.y ~= t1.y or t2.z ~= t1.z then
        flag = true  -- 動かせる
      end
    end
    if (not flag) then
      if i == 136 then
        flag = true  -- 動かせる
      else
        t2 = tiles[i + 1]
        if (not t2.visible) or t2.y ~= t1.y or t2.z ~= t1.z then
          flag = true  -- 動かせる
        end
      end
    end
    if flag then
      t1.movable = true  -- 仮に「取れる」とマークする
      count_movable[t1.id] = count_movable[t1.id] + 1  -- 動かせる牌をカウント
    end
  ::cont::
  end
  number_movable = 0
  number_remain = 0
  for i = 1, 136 do
    t1 = tiles[i]
    if t1.movable and count_movable[t1.id] < 2 then
      t1.movable = false  -- 同じ牌が１個しかないなら実は取れない
    end
    if t1.movable then number_movable = number_movable + 1 end
    if t1.visible then number_remain = number_remain + 1 end
    if t1.z > zmax then zmax = t1.z end
  end
  -- カーソル位置を設定
  if curs == 0 then curs = 1 end
  while curs <= 136 do
    t1 = tiles[indices[curs]]
    if t1.visible and t1.movable then break end
    curs = curs + 1
  end
  if curs > 136 then curs = 0 end
end

--  盤面を作成
function shuffle()
  local i, j, k, xx, yy, t
  j = 1
  --  牌をシャッフルする
  for i = 1, 136 do
    tiles[i] = math.floor((i + 3) / 4)
    indices[i] = i
  end
  love.math.setRandomSeed(seed)
  for i = 1, 200 do
    j = math.floor(love.math.random() * 136) + 1
    k = math.floor(love.math.random() * 136) + 1
    tiles[j], tiles[k] = tiles[k], tiles[j]
  end
  j = 1
  for i = 1, #layout do
    -- layout から (x, y, z) 座標を確定
    local ly = layout[i]
    local x1, y1, x2, y2, zz = ly[1], ly[2], ly[3], ly[4], ly[5]
    for yy = y1, y2 do
      for xx = x1, x2 do
        t = tiles[j]
        tiles[j] = { x = xx, y = yy, z = zz, id = t, movable = false, visible = true }
        j = j + 1
      end
    end
  end
  -- y, x, z の昇順にインデックスを並べ替え
  table.sort(indices, function(a, b)
    local t1, t2 = tiles[a], tiles[b]
    return (t1.y * 10000 + t1.x * 100 + t1.z) < (t2.y * 10000 + t2.x * 100 + t2.z)
  end)
  check_movable() -- 取れるかどうかを判定しておく
end

function love.load()
  local i
  --  画面サイズなど
  width = 800
  height = 600
  swidth, sheight, mode = love.window.getMode()
  mode.vsync = true
  love.window.setMode(swidth, sheight, mode)
  scale = math.min(width / swidth, height / sheight)
  transx = math.floor((swidth - width * scale) / 2)
  transy = math.floor((sheight - height * scale) / 2)
  love.window.setTitle("Mahjong Solitaire")
  -- テクスチャアトラス
  image = love.graphics.newImage("tiles.png")
  wid0, high0 = image:getDimensions()
  wid = 44
  high = 62
--  images = {}
  for i = 1, 34 do
    local x = (i - 1) % 9
    local y = math.floor((i - 1) / 9)
    quads[i] = love.graphics.newQuad(x * wid, y * high, wid, high, image:getDimensions())
--    images[i] = love.graphics.newImage(string.format("%02d.png", i))
  end
  love.graphics.setBackgroundColor(0, 0.31, 0)  -- 背景を濃い緑色に
  font = love.graphics.newFont(18)
  line = love.graphics.newText(font, "")
  shuffle()
end

function love.update(dt)
  love.timer.sleep(0.01)
end

function love.keypressed(key)
  local c, t1, t2
  c = curs
  if key == "r" then
    shuffle()
  elseif key == "n" then
    seed = seed + 1
    shuffle()
  end
  if c == 0 then return end
  if key == "right" then
    repeat
      c = c + 1
      if c == 137 then c = 1 end
      t1 = tiles[indices[c]]
      if t1.visible and t1.movable then break end
    until c == curs
    curs = c
  elseif key == "left" then
    repeat
      c = c - 1
      if c == 0 then c = 136 end
      t1 = tiles[indices[c]]
      if t1.visible and t1.movable then break end
    until c == curs
    curs = c
  elseif key == "return" then
    t1 = tiles[indices[curs]]
    if sel > 0 then
      -- 選択している牌がある
      t2 = tiles[indices[sel]]
      if t1.id == t2.id then
        -- 同じ牌を選んだ：消す
        t1.visible = false
        t2.visible = false
        sel = 0
        curs = 0
        check_movable() -- 取れる牌を調べ直す
      else
        -- 違う牌を選んだ：選択する牌を変更
        sel = curs
      end
    else
      -- この牌を選択
      sel = curs
    end
  end
end

function love.mousepressed(x, y, button, istouch)
  local ii
  x = (x - transx) / scale
  y = (y - transy) / scale
  for n = 136, 1, -1 do
    local t = tiles[indices[n]]
    local px = t.x * (wid - 2) + width / 2 - t.z * 3
    local py = t.y * (high - 8) + height / 2 - t.z * 7
    if t.movable then
      f = (x >= px and x < px + wid - 2 and y >= py and y < py + high - 8)
      if f then
        curs = n
        love.keypressed("return")
        break
      end
    end
  end
end

function love.draw()
  local px, py, t, n, cn, col, spn, flg
  love.graphics.translate(transx, transy)
  love.graphics.scale(scale, scale)
  love.graphics.setColor(0, 0.5, 0)
  love.graphics.rectangle("fill", 0, 0, width, 32)
  local s = string.format("Stage %3d      Remain %3d      Removable %3d      [R] Retry      [N] New", seed, number_remain, number_movable)
  line:set(s)
  love.graphics.setColor(1, 1, 1)
  love.graphics.draw(line, math.floor((width - line:getWidth()) / 2), 4)
  for n = 0, 135 do
    t = tiles[n + 1]
    if not t.visible then goto cont end
    px = t.x * (wid - 2) + width / 2 - t.z * 3
    py = t.y * (high - 8) + height / 2 - t.z * 7
    if sel > 0 and indices[sel] == n + 1 then
      love.graphics.setColor(1, 0, 0)
    elseif t.movable then
      love.graphics.setColor(1, 1, 0)
    else
      local b = 1 - (zmax - t.z) * 0.031
      love.graphics.setColor(b, b, b)
    end
    love.graphics.draw(image, quads[t.id], px, py)
    if curs > 0 and indices[curs] == n + 1 then
      love.graphics.setLineWidth(3)
      love.graphics.setColor(0, 0, 1)
      love.graphics.rectangle("line", px + 2.5, py + 2.5, wid - 8, high - 12)
    end
  ::cont::
  end
end
